/*
    Program to create a sparse file.
    Copyright (C) 2002 Bo Brantn.
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>

void PrintLastError(char* Prefix)
{
    LPVOID lpMsgBuf;

    FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        GetLastError(),
        0,
        (LPTSTR) &lpMsgBuf,
        0,
        NULL
        );

    fprintf(stderr, "%s: %s", Prefix, (LPTSTR) lpMsgBuf);

    LocalFree(lpMsgBuf);
}

int __cdecl main(int argc, char* argv[])
{
    LARGE_INTEGER   FileSize;
    HANDLE          FileHandle;
    DWORD           BytesReturned;

    if (argc < 2 || argc > 3)
    {
        fprintf(stderr, "syntax: mksparse <filename> [size[k|M|G]]\n");
        return -1;
    }

    if (argc == 3)
    {
        if (argv[2][strlen(argv[2]) - 1] == 'G')
        {
            FileSize.QuadPart = _atoi64(argv[2]) * 1024 * 1024 * 1024;
        }
        else if (argv[2][strlen(argv[2]) - 1] == 'M')
        {
            FileSize.QuadPart = _atoi64(argv[2]) * 1024 * 1024;
        }
        else if (argv[2][strlen(argv[2]) - 1] == 'k')
        {
            FileSize.QuadPart = _atoi64(argv[2]) * 1024;
        }
        else
        {
            FileSize.QuadPart = _atoi64(argv[2]);
        }
    }
    else
    {
        FileSize.QuadPart = 0;
    }

    FileHandle = CreateFile(
        argv[1],
        GENERIC_READ | GENERIC_WRITE,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL,
        CREATE_NEW,
        FILE_FLAG_NO_BUFFERING,
        NULL
        );

    if (FileHandle == INVALID_HANDLE_VALUE)
    {
        PrintLastError(argv[1]);
        return -1;
    }

    if (!DeviceIoControl(
        FileHandle,
        FSCTL_SET_SPARSE,
        NULL,
        0,
        NULL,
        0,
        &BytesReturned,
        NULL
        ))
    {
        PrintLastError(argv[1]);
        return -1;
    }

    if (!SetFilePointerEx(FileHandle, FileSize, 0, FILE_BEGIN))
    {
        PrintLastError(argv[1]);
        return -1;
    }

    if (!SetEndOfFile(FileHandle))
    {
        PrintLastError(argv[1]);
        return -1;
    }

    CloseHandle(FileHandle);

    return 0;
}
